#include "bsp.h"
#include "tty.h"

typedef union {
  byte b[4];
  long l;
  void *p;
} LongPtr;

static inline void *
long2ptr(long l)
{
  LongPtr lp;
  lp.l = l;
  return lp.p;
}
static inline void *
byte2ptr(byte h, byte m, byte l)
{
  LongPtr lp;
  lp.b[0] = l;
  lp.b[1] = m;
  lp.b[2] = h;
  lp.b[3] = 0;
  return lp.p;
}

#define F_IO_0	p9.b0
#define F_IO_1	p9.b1
#define F_IO_2	p9.b2
#define F_IO_3	p9.b3

extern void dllexit();

#define LA_ADDR 0x00100000L
volatile word *la = (word *)(LA_ADDR);

#define RAM_ADDR 0x00400000L
volatile word *ram = (word *)(RAM_ADDR);

static inline void
write_byte (long a, word d)
{
  ram[a] = d;
}

static inline word
read_byte (long a)
{
  return ram[a];
}

int phase = 0;

void do_phase (int delta)
{
#if 0
  while (delta > 0 && phase < 251)
    {
      F_IO_2 = 1;
      F_IO_2 = 0;
      delta --;
      phase ++;
    }
  while (delta < 0 && phase > -251)
    {
      F_IO_3 = 1;
      F_IO_3 = 0;
      delta ++;
      phase --;
    }
#endif
  phase += delta;

  phase &= 7;
  F_IO_1 = phase & 1;
  F_IO_2 = (phase & 2) >> 1;
  F_IO_3 = (phase & 4) >> 2;
}

//static word data_patterns[] = { 0x0001U, 0xfff3U, 0xaaaaU, 0x5555U };
static word data_patterns[] = { 0x55aaU, 0xaa55U, 0xaaaaU, 0x5555U };
static int pp=0;

static int
data_ok (volatile word *ram)
{
  int i;
  for (i=0; i<4; i++)
    {
      word p = data_patterns[pp++];
      word r;
      pp %= 4;
      ram[0] = 0xff00;
      ram[1] = p;
      ram[2] = 0x00ff;
      r = ram[1];
      if (r != p)
	{
	  la[0] = 1;
	  //cprintf("%X %w %w\n", ram, r, p);
	  return 0;
	}
    }
  return 1;
}

main()
{
  unsigned int i, j, t, u, done;
  long count, fails;
  int h, m, l;
  unsigned int oi, oj, op=-1;
  unsigned char new_pd9;
  long addr = 0;
  volatile word *r1, *r2;

  p10.b = 0x33;

  F_IO_0 = 1; /* This is RESET ACTIVE HIGH */
  F_IO_1 = 1; /* chip select ACTIVE LOW */
  F_IO_2 = 0; /* increment phase */
  F_IO_3 = 0; /* decrement phase */

  prcr.b = 15;

  pm0.b = 0x05;
  pm1.b = 0x00;
  cm0   = 0x00;
  ds = 0x0f;
  mcd = 0x12; /* divide by 1 */
  //mcd = 0x02; /* divide by 2 */
  //mcd = 0x04; /* divide by 4 */

  ewcr0 = 0x01;
  ewcr1 = 0x01;
  ewcr2 = 0x01;
  ewcr3 = 0x01;

  new_pd9 = pd9.b | 0x0f;
  prcr.b = 0x04;
  pd9.b = new_pd9; /* f_io_* */


  wait_ms (1);

#if 1
  while (tty_ready ())
    tty_readc ();
  cprintf("about to start...\n");
  while (!tty_ready ())
    ;
  tty_readc ();
#endif

  F_IO_0 = 0; /* de-assert reset */
  F_IO_1 = 0; /* chip select ACTIVE LOW */

  cprintf("running...\n");

  done = 0;
  phase = 0;
  do_phase (0);

#if 0
  cprintf("\033[1;1H\033[2J");
  ram[0] = 0xff00;
  ram[1] = 0x3333;
  ram[2] = 0x00ff;
  j = 0;
  count = 0;
  fails = 0;
  while (!done)
    {
      i = ram[0];
      if (tty_ready ())
	{
	  char ch = tty_readc ();
	  switch (ch)
	    {
	    case 'q':
	      goto done1;
	    case 'i': do_phase(+1); break;
	    case 'd': do_phase(-1); break;
	    }
	  count = 0;
	  fails = 0;
	}
      if (i != 0x3333)
	fails ++;
      count ++;
      if (i != j || phase != t)
	{
	  j = i;
	  t = phase;
	  cprintf("\033[1;1H phase %d  \n", phase);
	  cprintf("ram[0] %W\n", 0xff00);
	  cprintf("ram[1] %W\n", i);
	  cprintf("ram[2] %W\n", 0x00ff);
	}
      if ((count & 0xffff) == 0)
	cprintf("\r%X %X  ", count, fails);
    }
 done1:
#endif

#if 1
  cprintf("\033[1;1H\033[2J");
  while (!done)
    {
      if (tty_ready ())
	{
	  char ch = tty_readc ();
	  switch (ch)
	    {
	    case 'q':
	      goto done1;
	    case 'i': do_phase(+1); break;
	    case 'd': do_phase(-1); break;
	    }
	}
      cprintf("\033[1;1H delay %d \n", phase);
      for (h=0; h<=0xff; h += 16)
	{
	  cprintf("%b0000 : ", h);
	  for (m=0; m <= 0xff; m += 0x08)
	    {
	      int hb = h|(m>>4);

	      if (hb == 0)
		cprintf(" -");
	      else
		{
		  int ok;

		  ok = data_ok(byte2ptr(hb, m<<4, hb & 0xfe));
		  cprintf(" %d", ok);
		}
	    }
	  cprintf("\n");
	}
    }
 done1:
#endif

#if 0
  for (i=0; i<30*16*3; i++)
    ram[i+0x1000] = (i&1) ? 0xffff : 0x0000;
  for (i=0; i<32; i++)
    ram[i] = i;

  i = count = 0;
  done = 0;
  phase = 0;
  cprintf("\033[1;1H\033[2J");
  while (!done)
    {
      cprintf("\033[1;1H");
      cprintf("<< h < j  0  k > l >>  Quit\n");
      i = ram[2];
      j = ram[10];

      cprintf("  p %d 0002:%w 000a:%w    \r", phase, i, j);

      wait_ms (2);

      if (tty_ready())
	{
	  char ch = tty_readc ();
	  switch (ch)
	    {
	    case 'h':
	      do_phase (-10);
	      break;
	    case 'j':
	      do_phase (-1);
	      break;
	    case 'k':
	      do_phase (1);
	      break;
	    case 'l':
	      do_phase (10);
	      break;
	    case '0':
	      do_phase (-phase);
	      break;
	    case 'q':
	    case 13:
	    case 27:
	      done = 1;
	      break;
	    }
	}

      cprintf("\n\n");
      for (i=0; i<10; i++)
	{
	  for (j=0; j<16; j++)
	    {
	      word v;
	      t = 0x1000 + i*16 + j;
	      v = j&1 ? 0xffff : 0x0000;
	      u = ram[t];
	      if (v != u)
		{
		  cprintf(" \033[31m%w\033[0m", u);
		  //ram[t] = t;
		  ram[t] = v;
		}
	      else
		cprintf(" %w", u);
	    }
	  cprintf("\n");
	}
    }
  cprintf("\033[0m");

#endif

  prcr.b = 15;
  pm0.pm0mode = 0; /* single chip */
  prcr.b = 0;

}
